use std::collections::HashMap;
use std::env;
use std::fs::File;
use std::io;

use aoc2019::intcode;

#[derive(Hash, Default, PartialEq, Eq, Clone)]
struct Panel {
    x: isize,
    y: isize,
}

enum Color {
    Black,
    White,
}

impl From<isize> for Color {
    fn from(color: isize) -> Self {
        match color {
            0 => Color::Black,
            1 => Color::White,
            _ => panic!("invalid color"),
        }
    }
}

impl Into<isize> for &Color {
    fn into(self) -> isize {
        match self {
            Color::Black => 0,
            Color::White => 1,
        }
    }
}

impl Color {
    fn char(&self) -> char {
        match self {
            Color::Black => ' ',
            Color::White => '\u{2588}',
        }
    }
}

enum Direction {
    Up,
    Down,
    Left,
    Right,
}

enum Turn {
    Left,
    Right,
}

impl From<isize> for Turn {
    fn from(turn: isize) -> Self {
        match turn {
            0 => Turn::Left,
            1 => Turn::Right,
            _ => panic!("invalid turn direction"),
        }
    }
}

struct Robot {
    panel: Panel,
    direction: Direction,
}

impl Robot {
    fn turn(&mut self, direction: Turn) {
        self.direction = match self.direction {
            Direction::Up => match direction {
                Turn::Left => Direction::Left,
                Turn::Right => Direction::Right,
            },
            Direction::Right => match direction {
                Turn::Left => Direction::Up,
                Turn::Right => Direction::Down,
            },
            Direction::Down => match direction {
                Turn::Left => Direction::Right,
                Turn::Right => Direction::Left,
            },
            Direction::Left => match direction {
                Turn::Left => Direction::Down,
                Turn::Right => Direction::Up,
            },
        }
    }

    fn step(&mut self) {
        self.panel = match self.direction {
            Direction::Up => Panel {
                x: self.panel.x,
                y: self.panel.y + 1,
            },
            Direction::Left => Panel {
                x: self.panel.x + 1,
                y: self.panel.y,
            },
            Direction::Down => Panel {
                x: self.panel.x,
                y: self.panel.y - 1,
            },
            Direction::Right => Panel {
                x: self.panel.x - 1,
                y: self.panel.y,
            },
        }
    }
}

fn main() -> io::Result<()> {
    let filename = env::args().nth(1).expect("provide file as first param");
    let mut computer = intcode::Computer::from(File::open(filename)?);

    let mut canvas: HashMap<Panel, Color> = HashMap::new();
    canvas.insert(Panel { x: 0, y: 0 }, Color::White);
    let mut robot = Robot {
        panel: Panel::default(),
        direction: Direction::Up,
    };

    let mut input: isize = canvas
        .get(&robot.panel)
        .or_else(|| Some(&Color::Black))
        .unwrap()
        .into();
    while let Some(output) = computer.run_io(input) {
        canvas.insert(robot.panel.clone(), Color::from(output));

        let output = computer.run_must_output();
        robot.turn(Turn::from(output));
        robot.step();

        input = canvas
            .get(&robot.panel)
            .or_else(|| Some(&Color::Black))
            .unwrap()
            .into();
    }

    let mut xmin: isize = 0;
    let mut xmax: isize = 0;
    let mut ymin: isize = 0;
    let mut ymax: isize = 0;
    let painted = canvas
        .iter()
        .map(|(k, _)| {
            if k.x < xmin {
                xmin = k.x;
            }
            if k.x > xmax {
                xmax = k.x;
            }
            if k.y < ymin {
                ymin = k.y;
            }
            if k.y > ymax {
                ymax = k.y;
            }
        })
        .count();
    println!("painted: {}", painted);
    println!("canvas: x:{}-{} y:{}-{}", xmin, xmax, ymin, ymax);

    for y in (ymin..=ymax).rev() {
        for x in (xmin..=xmax).rev() {
            print!(
                "{}",
                canvas
                    .get(&Panel { x, y })
                    .unwrap_or_else(|| &Color::Black)
                    .char()
            )
        }
        println!()
    }

    Ok(())
}
